"""SQL Properties product




"""

__version__='$Revision: 1.3 $'[11:-2]

import traceback, sys, os
import Globals, App.Undo, socket, regex
from Products.ZSQLMethods.SQL import SQL, SQLConnectionIDs
from Globals import PersistentMapping
from Globals import Persistent
from Globals import HTMLFile, MessageDialog
from string import join,strip,split,lower
from App.Management import Navigation, Tabs
from Acquisition import Implicit, Explicit
import OFS.SimpleItem 
from OFS.Document import Document
from base64 import decodestring, encodestring
from urllib import quote, unquote
from ImageFile import ImageFile
import AccessControl.Role 
import string
from cPickle import loads, dumps

ListType=type([])

class DumpingMethods:
    def base64_load(self, v):
	return loads(decodestring(v))

    def base64_dump(self, v):
	return encodestring(dumps(v))[:-1] # chop \n

    def zopemarshal_load(self, v):
	raise "Unimplemented"

    def zopemarshal_dump(self, v):
	raise "Unimplemented"

class InstanceDict(DumpingMethods):
    """ """
    caching = 0

    def __init__(self, name, parent, mode="base64"):
	self.name = name
	self.parent = parent
	self.mode = mode
	if self.caching:
	    self.cachedict = {}
	self.InstanceDict = self

    def __len__(self):
 	rows = self.parent.sqlGetValue(instance_name=self.name)
        return len(rows)

    def __getitem__(self, name):
	"get"
	if self.caching and self.cachedict.has_key(name):
	    return self.cachedict[name]
	r = self.parent.sqlGetValue(instance_name=self.name, name=name)
	if len(r):
	    ret = self.base64_load(sqlattr(r[0], 'attr_value'))
	else:
	    raise KeyError
	if self.caching:
	    self.cachedict[name] = ret
	return ret


    def __setitem__(self, name, value):
	"setitem"
	encvalue = self.base64_dump(value)
	#print "encoded to", encvalue
	self.parent.sqlSetValue(instance_name=self.name,
				    name=name, value=encvalue)
	if self.caching:
	    self.cachedict[name] = value
	return None

    set = __setitem__

    InstanceSet = __setitem__

    def keys(self):
	rows = self.parent.sqlGetValue(instance_name=self.name)
	l = []
	for r in rows:
	   l.append(sqlattr(r, 'attr_name'))
	return l

    def values(self):
	rows = self.parent.sqlGetValue(instance_name=self.name)
	l = []
	for r in rows:
	    v = self.base64_load(sqlattr(r, 'attr_value'))
	    l.append(v) 
	    if self.caching:
		self.cachedict[r] = v
	return l

    def items(self):
	rows = self.parent.sqlGetValue(instance_name=self.name)
	l = []
	for r in rows:
	    name = sqlattr(r, 'attr_name')
	    #print name
	    value = self.base64_load(sqlattr(r, 'attr_value'))
	    l.append(( name, value ))
	    if self.caching:
		self.cachedict[name] = value
	return l

    def has_key(self, name):
	#print "has_key called"
	r = self.parent.sqlGetValue(instance_name=self.name,
					    name=name)
	if len(r):
	    return 1
	else:
	    return 0

    def __str__(self):
	"string repr"
	l = ["Property Variables:\n"]
	for n,v in self.items():
	    l.append("%s: %s"%(n,str(v)))
	return join(l, '\n')

    dump = __str__

class InstanceObj(Implicit):
    def __init__(self, name, parent, mode="base64"):
	self.dict = InstanceDict(name, parent, mode)

    def InstanceSet(self, name, value):
	self.dict[name] = value

    def InstanceGet(self, name):
	return self.dict[name]

    def __getattr__(self, name):
        if self.__dict__.has_key(name):
            return self.__dict__[name]
        if self.dict.has_key(name):
            return self.dict[name]
        else:
            raise AttributeError, name

class PersistentInstanceObj(OFS.SimpleItem.Item, Persistent, Implicit,
                      AccessControl.Role.RoleManager, InstanceObj):
    caching = 1
    def __init__(self, name, parent, mode="base64"):
	self.name = name
	self.parent = parent
	self.mode = mode
	if self.caching:
	    self.cachedict = {}


class SQLProperties(OFS.SimpleItem.Item, Persistent, Implicit, 
		      AccessControl.Role.RoleManager, DumpingMethods):
    """ An SQL Properties. """


    meta_type='SQL Properties'
    id       ='Properties'
    title    ='SQL Properties'
    icon     ='misc_/SQLProperties/SQLProperties_icon'

    manage_options=(
    {'label':'Instances',  'action':'manage_instances'},
    {'label':'Attributes', 'action':'manage_attributes'},
    {'label':'Properties', 'action':'manage_main'},
    {'label':'Security',   'action':'manage_access'},
    )

    __ac_permissions__=(
    ('View management screens',
     ['manage','manage_menu','manage_main','manage_tabs', 'manage_attributes','manage_instances' ]),
    ('Change SQLProperties',     ['manage_edit', 'newAttr' ]),
    ('View', ['__call__','']),
    )

    manage_instances=HTMLFile('manage_instances', globals())
    manage_attributes=HTMLFile('manage_attributes', globals())
    manage_main=HTMLFile('main', globals())

    def __init__(self, conn_id, schema_name, encode_mode):
        self.conn_id=conn_id
	self.schema_name = schema_name
	self.encode_mode = encode_mode
	self.conn_id = conn_id
        
        self.sqlDelAttr=SQL('sqlDelAttr', '', conn_id,
		  'schema_name="%s" name'%(self.schema_name),
		  loadsql('del_attr.sql'))

        self.sqlNewAttr=SQL('sqlNewAttr', '', conn_id,
		  'schema_name="%s" name value'%(self.schema_name),
		  loadsql('new_attr.sql'))

        self.sqlDelInstance=SQL('sqlDelInstance', '', conn_id,
		  'schema_name="%s" instance_name'%(self.schema_name),
		  loadsql('del_instance.sql'))

        self.sqlNewInstance=SQL('sqlNewInstance', '', conn_id,
		  'schema_name="%s" instance_name'%(self.schema_name),
		  loadsql('new_instance.sql'))

        self.sqlGetValue=SQL('sqlGetValue', '', conn_id,
		  'schema_name="%s" instance_name name=""'%(self.schema_name),
		  loadsql('get_value.sql'))

        self.sqlSetValue=SQL('sqlSetValue', '', conn_id,
		 'schema_name="%s" instance_name name value'%(self.schema_name),
		  loadsql('set_value.sql'))

        self.sqlListAttributes=SQL('sqlListAttributes', '', conn_id,
                             'schema_name="%s"'%(self.schema_name),
                              loadsql('list_attributes.sql'))

        self.sqlListInstances=SQL('sqlListInstances', '', conn_id,
                             'schema_name="%s"'%(self.schema_name),
                              loadsql('list_instances.sql'))

        self.sqlNewSchema=SQL('sqlNewSchema', '', conn_id,
                             'schema_name="%s"'%(self.schema_name),
                              loadsql('new_schema.sql'))

    def instanceIds(self):
	"instances..."
	l = self.sqlListInstances()
	return map(lambda x:x.instance_id, l)

    def instanceValues(self):
	"instances..."
	l = self.sqlListInstances()
	return map(lambda x:x.instance_name, l)

    def instanceItems(self):
	"instances..."
	l = self.sqlListInstances()
	return l

    def attributeIds(self):
	"attributes..."
	l = self.sqlListAttributes()
	return map(lambda x:x.attr_name, l)

    def attributeValues(self):
	"attributes..."
	l = self.sqlListAttributes()
	return map(lambda x:x.attr_value, l)

    def attributeItems(self):
	"attributes..."
	l = self.sqlListAttributes()
	return l

    def enumQueries(self):
        list=[]
        type=SQL.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
	list.sort()
        return list

    def enumDocuments(self):
        list=[]
        type=Document.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
        return list

    def newInstance(self, name, REQUEST=None):
	"create a new instance"
        self.sqlNewInstance(instance_name=name)
        if REQUEST:
            return MessageDialog(
                   title  ='new instance created',
                   message='instance %s has been created'%name,
                   action =REQUEST['URL1']+'/manage_instances',
                   target ='manage_main')

    def newAttr(self, name, value, REQUEST=None):
	"set new attribute"
	encvalue = self.base64_dump(value)
	#print "encoded to", encvalue
	self.sqlNewAttr(name=name, value=encvalue)
        if REQUEST:
            return MessageDialog(
                   title  ='new attribute created',
                   message='attribute %s has been created'%name,
                   action =REQUEST['URL1']+'/manage_attributes',
                   target ='manage_main')

    def manage_edit(self, connection_id, encode_mode, REQUEST=None):
	"""Change properties"""
        if connection_id != self.conn_id:
            # Connection changed - update sql methods!
            self.conn_id=connection_id
            for ob in self.enumQueries():
                title=ob.title
                args=ob.arguments_src
                src=ob.src
                ob.manage_edit(title, self.conn_id, args, src)
	self.encode_mode = encode_mode
        if REQUEST:
            return MessageDialog(
                   title  ='SQLProperties Changed',
                   message='SQLProperties properties have been updated',
                   action =REQUEST['URL1']+'/manage_main',
                   target ='manage_main')

    def __bobo_traverse__(self, REQUEST, name=None):
        if hasattr(self,name):
            return getattr(self,name)
	else:
	    return InstanceObj(name, self, self.encode_mode).__of__(self)


manage_addSQLPropertiesForm=HTMLFile('addSQLProperties', globals())

def manage_addSQLProperties(self, id, schema_name, connection_id, 
				encode_mode, REQUEST=None):
    """ """
    if hasattr(self.aq_base, id):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains an item called %s'%id,
               action ='%s/manage_main' % REQUEST['PARENT_URL'])
    ob=SQLProperties(connection_id, schema_name, encode_mode)
    self._setObject(id, ob)
    newob = getattr(self, id)
    newob.sqlNewSchema()
    if REQUEST: return self.manage_main(self,REQUEST)

def loadsql(filename):
    import os
    #print os.getcwd()
    try:
	fp = open("dbspecific/db/%s"%filename)
    except:
	fp = open("lib/python/Products/SQLProperties/dbspecific/db/%s"%filename)
    #print fp
    return fp.read()

# A getattr-like helper that hopefully can compensate for the
# various oddities in data returned by different stupidbases.

import Missing
mt=type(Missing.Value)

def typeconv(val):
    if type(val)==mt:
        return ''
    return val

from string import upper, lower

def sqlattr(ob, attr):
    name=attr
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    attr=upper(attr)
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    attr=lower(attr)
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    raise NameError, name

def absattr(attr):
    if callable(attr): return attr()
    return attr

def reqattr(request, attr):
    try:    return request[attr]
    except: return None

